package App::Todo;

our $VERSION = 0.96;

=head1 NAME

App::Todo - Provides todo.pl, a command-line interface to Hiveminder

=head1 DESCRIPTION

Please see the documentation for L<todo.pl>.

=cut

1;
